<?php declare(strict_types=1);

namespace Lean;

use Dotenv\Dotenv;

use DI\ContainerBuilder,
    DI\Container;

use League\Route\Strategy\ApplicationStrategy;

use Psr\Http\Message\ServerRequestInterface;

use Tell\I18n;
use Ulmus\Container\AdapterProxy;

use Laminas\Diactoros\ServerRequestFactory,
    Laminas\HttpHandlerRunner\Emitter\EmitterInterface;

use Ulmus\Ulmus;

class Kernel {

    protected Container $container;

    protected string $locale;

    public array $paths = [];

    public int $errorReporting = E_ALL & ~E_USER_DEPRECATED & ~E_DEPRECATED & ~E_STRICT & ~E_NOTICE;

    public array $definitionFilePaths;

    public string $errorLogPath;

    public string $projectPath;

    public function __construct(string $projectPath)
    {
        $this->projectPath = $projectPath;

        $this->setEnvironment();
        $this->initializeEngine();
        $this->setupContainer();
        $this->serviceContainer();
        $this->handleRequest();
    }

    public static function putenv($var, $value)
    {
        putenv("$var=$value");
        $_ENV[$var] = $value;
    }

    protected function setEnvironment() {
        static::putenv('PROJECT_PATH', $this->projectPath);

        // Environment vars (accessible from \DI\env(), getenv(), $_ENV and $_SERVER)
        Dotenv::create(getenv("PROJECT_PATH"))->load();

        // Override using headers
        if ( ( $keys = getenv('KEYS') ) && ( $auth = $_SERVER["HTTP_X_DEV_AUTH"] ?? null ) && in_array($auth, explode(',', $keys)) ) {
            foreach (['APP_ENV', 'DEBUG',] as $env) {
                if (null !== $value = $_SERVER["HTTP_X_$env"] ?? null) {
                    static::putenv($env, $value);
                }
            }
        }

        // Paths and directories
        foreach($this->paths as $name => $envkey) {
            if ( ! getenv($name) ) {
                $path = getenv("PROJECT_PATH") . DIRECTORY_SEPARATOR . getenv($envkey);

                if (getenv('DEBUG') && ! file_exists($path)) {
                    if (false === mkdir($path, 0755, true)) {
                        throw new \Exception(sprintf("Folder '%s' could not be created", $path));
                    }
                }

                static::putenv($name, realpath($path));
            }
        }
    }

    protected function initializeEngine() : self
    {
        # Defining default locale and timezone
        date_default_timezone_set(getenv("DEFAULT_TIMEZONE"));

        setlocale(LC_ALL, $this->locale = getenv("DEFAULT_LOCAL"));
        setlocale(LC_TIME, getenv("DEFAULT_TIME"), getenv("DEFAULT_TIME_FALLBACK"));

        if ( class_exists('Locale') ) {
            \Locale::setDefault($this->locale);
        }

        ini_set("log_errors", "1");
        ini_set("error_log", $this->errorLogPath);
        ini_set('display_errors', getenv("DEBUG") ? 'on' : 'on');

        error_reporting($this->errorReporting);

        return $this;
    }

    protected function setupContainer() : self
    {
        $containerBuilder = new ContainerBuilder();

        $containerBuilder->useAttributes(true);

        if (getenv("APP_ENV") === "prod") {
            if (getenv("CACHE_PATH")) {
                # check if ACPU is there first, $containerBuilder->enableDefinitionCache();

                $containerBuilder->enableCompilation(getenv("CACHE_PATH") . "/di/");
                # No proxies yet... $containerBuilder->writeProxiesToFile(true, getenv("CACHE_PATH") . "/di/");
            }
        }
        else {
            $compiled = getenv("CACHE_PATH") . "/di/CompiledContainer.php";

            if ( file_exists($compiled) ) {
                unlink($compiled);
            }
        }

        foreach(Lean::getDefinitionsPathsFromComposer() as $path) {
            $containerBuilder->addDefinitions($path);
        }

        foreach($this->definitionFilePaths ?? [] as $path) {
            $containerBuilder->addDefinitions($path);
        }

        $containerBuilder->addDefinitions([
            'config' => function () {
                return array_merge_recursive(
                    Lean::autoloadConfigFromComposerExtra(), []
                );
            }
        ]);

        $this->container = $containerBuilder->build();

        return $this;
    }

    protected function serviceContainer() : self
    {
        if ($this->container->has('lean.autoload')) {
            foreach($this->container->get('lean.autoload') as $class) {
                $this->container->get($class);
            }
        }

        # Must be removed from KERNEL !
        $this->container->has('ulmus.caching') and ( Ulmus::$cache = $this->container->get('ulmus.caching') );
        # $this->container->has(AdapterProxy::class) and ( $ )

        return $this;
    }

    protected function handleRequest() : self
    {
        ServerRequestFactory::fromGlobals($_SERVER, $_GET, $_POST, $_COOKIE, $_FILES);

        $routeFactory = $this->container->get(Routing\RouteDefinitionInterface::class);

        $router = $routeFactory($this->container);

        $router->setStrategy($this->container->get(ApplicationStrategy::class));

        // Handle requests
        $response = $router->dispatch($this->container->get(ServerRequestInterface::class));

        // Reply to clients
        $this->container->get(EmitterInterface::class)->emit($response);

        return $this;
    }
};
